import torch
import torch.nn as nn
import torch.nn.functional as F
from joetorch.nn.modules import ConvResidualBlock, ConvAttentionBlock

class LargeEncoder(nn.Sequential):
    def __init__(self, in_channels: int, out_channels: int, hidden_channels: int=512):

        in_c = in_channels
        out_c = out_channels
        h_c = hidden_channels

        super().__init__(
            # (Batch_Size, Channel, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            nn.Conv2d(in_c, h_c//4, kernel_size=3, padding=1),
            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            ConvResidualBlock(h_c//4, h_c//4),
            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            ConvResidualBlock(h_c//4, h_c//4),

            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height/2, Width/2)
            nn.Conv2d(h_c//4, h_c//4, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c//4, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/2, Width/2)
            ConvResidualBlock(h_c//4, h_c//2),
            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/2, Width/2)
            ConvResidualBlock(h_c//2, h_c//2),

            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/4, Width/4)
            nn.Conv2d(h_c//2, h_c//2, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c//2, Height/4, Width/4) -> (Batch_Size, h_c, Height/4, Width/4)
            ConvResidualBlock(h_c//2, h_c),
            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/4, Width/4)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.Conv2d(h_c, h_c, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvAttentionBlock(h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.GroupNorm(32, h_c),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.SiLU(),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, out_c, Height/8, Width/8)
            nn.Conv2d(h_c, out_c, kernel_size=3, padding=1),
            # (Batch_Size, out_c, Height/8, Width/8) -> (Batch_Size, out_c, Height/8, Width/8)
            nn.Conv2d(out_c, out_c, kernel_size=1, padding=0)
            
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, Channel, Height, Width)

        for module in self:
            if getattr(module, 'stride', None) == (2,2):
                # (Padding_Left, Padding_Right, Padding_Top, Padding_Bottom)
                x = F.pad(x, (0, 1, 0, 1))
            x = module(x)

        x *= 0.18215

        return x


class MediumEncoder(nn.Sequential):
    def __init__(self, in_channels: int, out_channels: int, hidden_channels: int=256):

        in_c = in_channels
        out_c = out_channels
        h_c = hidden_channels

        super().__init__(
            # (Batch_Size, Channel, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            nn.Conv2d(in_c, h_c//4, kernel_size=3, padding=1),
            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            ConvResidualBlock(h_c//4, h_c//4),

            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height/2, Width/2)
            nn.Conv2d(h_c//4, h_c//4, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c//4, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/2, Width/2)
            ConvResidualBlock(h_c//4, h_c//2),

            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/4, Width/4)
            nn.Conv2d(h_c//2, h_c//2, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c//2, Height/4, Width/4) -> (Batch_Size, h_c, Height/4, Width/4)
            ConvResidualBlock(h_c//2, h_c),

            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.Conv2d(h_c, h_c, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvAttentionBlock(h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.GroupNorm(32, h_c),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.SiLU(),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, out_c, Height/8, Width/8)
            nn.Conv2d(h_c, out_c, kernel_size=3, padding=1),
            # (Batch_Size, out_c, Height/8, Width/8) -> (Batch_Size, out_c, Height/8, Width/8)
            nn.Conv2d(out_c, out_c, kernel_size=1, padding=0)
            
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, Channel, Height, Width)

        for module in self:
            if getattr(module, 'stride', None) == (2,2):
                # (Padding_Left, Padding_Right, Padding_Top, Padding_Bottom)
                x = F.pad(x, (0, 1, 0, 1))
            x = module(x)

        x *= 0.18215

        return x


class SmallEncoder(nn.Sequential):
    def __init__(self, in_channels: int, out_channels: int, hidden_channels: int=128):

        in_c = in_channels
        out_c = out_channels
        h_c = hidden_channels

        super().__init__(
            # (Batch_Size, Channel, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            nn.Conv2d(in_c, h_c//4, kernel_size=3, padding=1),
            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height, Width)
            ConvResidualBlock(h_c//4, h_c//4),

            # (Batch_Size, h_c//4, Height, Width) -> (Batch_Size, h_c//4, Height/2, Width/2)
            nn.Conv2d(h_c//4, h_c//4, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c//4, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/2, Width/2)
            ConvResidualBlock(h_c//4, h_c//2),

            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height/4, Width/4)
            nn.Conv2d(h_c//2, h_c//2, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c//2, Height/4, Width/4) -> (Batch_Size, h_c, Height/4, Width/4)
            ConvResidualBlock(h_c//2, h_c),

            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.Conv2d(h_c, h_c, kernel_size=3, stride=2, padding=0),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.GroupNorm(32, h_c),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            nn.SiLU(),
            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, out_c, Height/8, Width/8)
            nn.Conv2d(h_c, out_c, kernel_size=3, padding=1),
            # (Batch_Size, out_c, Height/8, Width/8) -> (Batch_Size, out_c, Height/8, Width/8)
            nn.Conv2d(out_c, out_c, kernel_size=1, padding=0)
            
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, Channel, Height, Width)

        for module in self:
            if getattr(module, 'stride', None) == (2,2):
                # (Padding_Left, Padding_Right, Padding_Top, Padding_Bottom)
                x = F.pad(x, (0, 1, 0, 1))
            x = module(x)

        x *= 0.18215

        return x

class LargeDecoder(nn.Sequential):
    def __init__(self, in_channels: int, out_channels: int, hidden_channels: int=512):
        in_c = in_channels
        out_c = out_channels
        h_c = hidden_channels

        super().__init__(

            nn.Conv2d(in_c, in_c, kernel_size=1, padding=0),
            nn.Conv2d(in_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c),
            ConvAttentionBlock(h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/4, Width/4)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/2, Width/2)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c//2),
            ConvResidualBlock(h_c//2, h_c//2),
            ConvResidualBlock(h_c//2, h_c//2),

            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height, Width)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c//2, h_c//2, kernel_size=3, padding=1),
            ConvResidualBlock(h_c//2, h_c//4),
            ConvResidualBlock(h_c//4, h_c//4),
            ConvResidualBlock(h_c//4, h_c//4),
            nn.GroupNorm(32, h_c//4),
            nn.SiLU(),

            # (Batch_Size, h//4, Height, Width) -> (Batch_Size, out_channels, Height, Width)
            nn.Conv2d(h_c//4, out_c, kernel_size=3, padding=1),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, 4, Height/8, Width/8)

        x /= 0.18215

        for module in self:
            x = module(x)

        # (Batch_Size, 3, Height, Width)
        return x

class MediumDecoder(nn.Sequential):

    def __init__(self, in_channels: int, out_channels: int, hidden_channels: int=256):
        in_c = in_channels
        out_c = out_channels
        h_c = hidden_channels

        super().__init__(
            nn.Conv2d(in_c, in_c, kernel_size=1, padding=0),
            nn.Conv2d(in_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c),
            ConvAttentionBlock(h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/4, Width/4)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c),
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/2, Width/2)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c//2),
            ConvResidualBlock(h_c//2, h_c//2),

            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height, Width)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c//2, h_c//2, kernel_size=3, padding=1),
            ConvResidualBlock(h_c//2, h_c//4),
            ConvResidualBlock(h_c//4, h_c//4),
            nn.GroupNorm(32, h_c//4),
            nn.SiLU(),

            # (Batch_Size, h//4, Height, Width) -> (Batch_Size, out_channels, Height, Width)
            nn.Conv2d(h_c//4, out_c, kernel_size=3, padding=1),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, 4, Height/8, Width/8)

        x /= 0.18215

        for module in self:
            x = module(x)

        # (Batch_Size, 3, Height, Width)
        return x

class SmallDecoder(nn.Sequential):

    def __init__(self, in_channels: int, out_channels: int, hidden_channels: int=128):
        in_c = in_channels
        out_c = out_channels
        h_c = hidden_channels

        super().__init__(

            nn.Conv2d(in_c, in_c, kernel_size=1, padding=0),
            nn.Conv2d(in_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c),
            ConvAttentionBlock(h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/8, Width/8)
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/8, Width/8) -> (Batch_Size, h_c, Height/4, Width/4)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c),

            # (Batch_Size, h_c, Height/4, Width/4) -> (Batch_Size, h_c, Height/2, Width/2)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c, h_c, kernel_size=3, padding=1),
            ConvResidualBlock(h_c, h_c//2),

            # (Batch_Size, h_c//2, Height/2, Width/2) -> (Batch_Size, h_c//2, Height, Width)
            nn.Upsample(scale_factor=2),
            nn.Conv2d(h_c//2, h_c//2, kernel_size=3, padding=1),
            ConvResidualBlock(h_c//2, h_c//4),
            nn.GroupNorm(32, h_c//4),
            nn.SiLU(),

            # (Batch_Size, h//4, Height, Width) -> (Batch_Size, out_channels, Height, Width)
            nn.Conv2d(h_c//4, out_c, kernel_size=3, padding=1),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, 4, Height/8, Width/8)

        x /= 0.18215

        for module in self:
            x = module(x)

        # (Batch_Size, 3, Height, Width)
        return x


from joetorch.nn.modules import *
from joetorch.nn.mnist import *
from joetorch.nn.convmixer import *
from joetorch.nn.stable_diffusion import *

__all__ = ['MLP', 'EncBlock', 'DecBlock', 'ConvResidualBlock', 'ConvMixer', 'SelfAttention', 'CrossAttention','LargeEncoder', 'MediumEncoder', 'SmallEncoder', 'LargeDecoder', 'MediumDecoder', 'SmallDecoder', 'MNIST_MLP_Encoder', 'MNIST_MLP_Decoder', 'MNIST_CNN_Encoder', 'MNIST_CNN_Decoder', 'MNIST_AE']



import torch.nn as nn

# Adapted from https://github.com/locuslab/convmixer/blob/main/convmixer.py
class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x):
        return self.fn(x) + x

# defaults for 128x128 images
class ConvMixer(nn.Module):
    def __init__(self, in_dim, out_dim, depth, kernel_size=5, patch_size=4):
        super().__init__()
        self.net = nn.Sequential(
        nn.Conv2d(in_dim, out_dim, kernel_size=patch_size, stride=patch_size),
        nn.GELU(),
        nn.BatchNorm2d(out_dim),
        *[nn.Sequential(
                Residual(nn.Sequential(
                    nn.Conv2d(out_dim, out_dim, kernel_size, groups=out_dim, padding="same"),
                    nn.GELU(),
                    nn.BatchNorm2d(out_dim)
                )),
                nn.Conv2d(out_dim, out_dim, kernel_size=1),
                nn.GELU(),
                nn.BatchNorm2d(out_dim)
        ) for i in range(depth)],
        nn.AdaptiveAvgPool2d((1,1)),
        nn.Flatten(),
        )

    def forward(self, x, stop_at=None):
        if stop_at == 0:
            return x
        else: 
            return self.net(x)
        


import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from torch.nn.functional import scaled_dot_product_attention

class SelfAttention(nn.Module):

    def __init__(self, n_heads: int, d_embed: int, in_proj_bias: bool = True, out_proj_bias: bool = True):
        super().__init__()

        assert d_embed % n_heads == 0, f"d_embed {d_embed} must be divisible by n_heads {n_heads}"

        self.n_heads = n_heads
        self.d_head = d_embed // n_heads

        self.in_proj = nn.Linear(d_embed, 3 * d_embed, bias=in_proj_bias)
        self.out_proj = nn.Linear(d_embed, d_embed, bias=out_proj_bias)

    def forward(self, x: torch.Tensor, causal_mask: bool = False) -> torch.Tensor:
        # x: (Batch_size, Seq_Len, Dim)

        input_shape = x.shape
        batch_size, seq_len, d_embed = input_shape        
        interim_shape = (batch_size, seq_len, self.n_heads, self.d_head)

        # (Batch_size, Seq_Len, Dim) -> (Batch_size, Seq_Len, 3*Dim) -> 3 x (Batch_size, Seq_Len, Dim)
        q, k, v = self.in_proj(x).chunk(3, dim=-1) 

        # (Batch_size, Seq_Len, Dim) -> (Batch_size, Seq_Len, n_heads, d_head) -> (Batch_size, n_heads, Seq_Len, d_head)
        q = q.view(*interim_shape).permute(0, 2, 1, 3).contiguous()
        k = k.view(*interim_shape).permute(0, 2, 1, 3).contiguous()
        v = v.view(*interim_shape).permute(0, 2, 1, 3).contiguous()

        # Use flash attention via scaled_dot_product_attention
        output = scaled_dot_product_attention(
            q, k, v,
            attn_mask=None,
            dropout_p=0.0,
            is_causal=causal_mask
        )

        # (Batch_size, n_heads, Seq_Len, d_head) -> (Batch_size, Seq_Len, n_heads, d_head) -> (Batch_size, Seq_Len, Dim)
        output = output.permute(0, 2, 1, 3).contiguous().view(*input_shape)

        output = self.out_proj(output)

        # (Batch_size, Seq_Len, Dim)
        return output


class CrossAttention(nn.Module):
    
    def __init__(self, n_heads: int, d_embed: int, d_cross: int, in_proj_bias: bool = True, out_proj_bias: bool = True):
        super().__init__()

        assert d_embed % n_heads == 0, f"d_embed {d_embed} must be divisible by n_heads {n_heads}"

        self.n_heads = n_heads
        self.d_head = d_embed // n_heads

        self.q_proj = nn.Linear(d_embed, d_embed, bias=in_proj_bias)
        self.kv_proj = nn.Linear(d_cross, d_embed*2, bias=in_proj_bias)
        self.out_proj = nn.Linear(d_embed, d_embed, bias=out_proj_bias)
    
    def forward(self, query, keys_n_values):
        # query: (Batch_size, Seq_Len_Q, Dim_Q)
        # keys_n_values: (Batch_size, Seq_Len_KV, Dim_KV) = (Batch_Size, 77, 768)

        input_shape = query.shape
        batch_size, sequence_length, d_embed = input_shape
        interim_shape = (batch_size, -1, self.n_heads, self.d_head)

        # Project inputs
        q = self.q_proj(query)
        kv = self.kv_proj(keys_n_values)
        k, v = kv.chunk(2, dim=-1)

        q = q.view(*interim_shape).permute(0, 2, 1, 3).contiguous()
        k = k.view(*interim_shape).permute(0, 2, 1, 3).contiguous()
        v = v.view(*interim_shape).permute(0, 2, 1, 3).contiguous()

        # Use flash attention via scaled_dot_product_attention
        output = scaled_dot_product_attention(
            q, k, v,
            attn_mask=None,
            dropout_p=0.0,
            is_causal=False
        )

        output = output.permute(0, 2, 1, 3).contiguous().view(*input_shape)
        output = self.out_proj(output)
        return output
